from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple, List, Callable

import js
from js import document, console, ResizeObserver
from pyodide.ffi import create_proxy
from pyodide.ffi.wrappers import add_event_listener, remove_event_listener

import wwwpy.remote.component as wpc
from wwwpy.remote import dict_to_js
from typing import NamedTuple
import logging

logger = logging.getLogger(__name__)


class Geometry(NamedTuple):
    top: int
    left: int
    width: int
    height: int


class DraggableComponent(wpc.Component, tag_name='wwwpy-window'):
    window_div: wpc.HTMLElement = wpc.element()
    window_title_div: wpc.HTMLElement = wpc.element()
    client_x = 0
    client_y = 0
    css_border = 2  # 1px border on each side, so we need to subtract 2px from width and height
    geometry_change_listeners: List[Callable[[], None]] = []

    def root_element(self):
        return self.shadow

    def init_component(self):
        self.shadow = self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.shadow.innerHTML = """
<style>
.window {
  z-index: 100000;  
  background-color: black;
  border: 1px solid #d3d3d3;
  resize: both;  
  overflow: hidden;
  position: absolute;  /* Changed from 'relative' to 'absolute' */
  display: flex;
  flex-direction: column;
}

.window-title {
  padding: 10px;
  cursor: move;
  z-index: 1001;
  background-color: #2196F3;
  color: #fff;
}

.window-body {
  overflow: auto;
}
</style>        
<div data-name="window_div" class='window'>
    <div  data-name="window_title_div" class='window-title' >
        <slot name='title' >slot=title</slot>
    </div>
   <div class='window-body'>
        <slot>slot=default</slot>
    </div>    
</div> 
"""
        self.client_x = 0
        self.client_y = 0
        tb = self.window_div

        def tb_print(*args):
            console.log(f'offsets of window_div: {self.geometry()}')

        tb.print = tb_print

        def on_resize(entries, observer):
            tb.print()
            self._on_geometry_change()

        resize_observer = ResizeObserver.new(create_proxy(on_resize))
        resize_observer.observe(self.window_div)

    def _on_geometry_change(self):
        for listener in self.geometry_change_listeners:
            listener()

    def window_title_div__touchstart(self, e: js.TouchEvent):
        self._move_start(e)

    def window_title_div__mousedown(self, e: js.MouseEvent):
        self._move_start(e)

    def _move_start(self, e: js.MouseEvent | js.TouchEvent):
        e.preventDefault()
        self.client_x = clientX(e)  # e.clientX
        self.client_y = clientY(e)  # e.clientY
        add_event_listener(document, 'mousemove', self._move)
        add_event_listener(document, 'mouseup', self._move_stop)
        add_event_listener(document, 'touchmove', self._move)
        add_event_listener(document, 'touchend', self._move_stop)

    def _move(self, event: js.MouseEvent | js.TouchEvent):
        x = clientX(event)
        y = clientY(event)
        delta_x = self.client_x - x
        delta_y = self.client_y - y
        self.client_x = x
        self.client_y = y

        # Get current 'left' and 'top' from style, defaulting to 0 if not set
        current_left = float(self.window_div.style.left.rstrip('px')) if self.window_div.style.left else 0
        current_top = float(self.window_div.style.top.rstrip('px')) if self.window_div.style.top else 0

        new_left = current_left - delta_x
        new_top = current_top - delta_y

        self.set_position(f'{new_left}px', f'{new_top}px')
        self._on_geometry_change()

    def _move_stop(self, event):
        remove_event_listener(document, 'mousemove', self._move)
        remove_event_listener(document, 'mouseup', self._move_stop)
        remove_event_listener(document, 'touchmove', self._move)
        remove_event_listener(document, 'touchend', self._move_stop)

    def geometry(self) -> Geometry:
        t = self.window_div
        res = Geometry(
            t.offsetTop,
            t.offsetLeft,
            t.offsetWidth - self.css_border,
            t.offsetHeight - self.css_border
        )
        return res

    def set_geometry(self, geometry_tuple):
        top, left, width, height = geometry_tuple
        self.set_position(f"{left}px", f"{top}px")
        self.set_size(f"{height}px", f"{width}px")

    def set_position(self, left: str | None = None, top: str | None = None):
        if top:
            logger.info(f'set_position: top={top}')
            top_check = float(top.removesuffix('px'))
            if top_check < 0:
                top = '0px'
            self.window_div.style.top = top
        if left:
            g = self.geometry()
            left_check = float(left.removesuffix('px'))
            if (left_check + g.width) > 30:
                self.window_div.style.left = left

    def set_size(self, height: str | None = None, width: str | None = None):
        if height:
            self.window_div.style.height = height
        if width:
            self.window_div.style.width = width

    def acceptable_geometry(self) -> bool:
        g = self.geometry()
        return g.width > 100 and g.height > 100 and g.top > 0 and g.left > 0


def clientX(event: js.MouseEvent | js.TouchEvent):
    #   var top = e.clientY || e.targetTouches[0].pageY;
    # return event.clientX if hasattr(event, 'clientX') else event.targetTouches[0].clientX
    if hasattr(event, 'targetTouches'):
        return list(event.targetTouches)[0].clientX
    return event.clientX


def clientY(event: js.MouseEvent | js.TouchEvent):
    return event.clientY if hasattr(event, 'clientY') else list(event.targetTouches)[0].clientY


@dataclass
class WindowResult:
    window: DraggableComponent

    @property
    def element(self):
        return self.window.element


def new_window(title: str, closable=True) -> WindowResult:
    win = DraggableComponent()
    # language=html
    ct = ClosableTitle()
    ct.element.setAttribute('slot', 'title')
    ct.title.innerHTML = title
    ct.close.onclick = lambda ev: win.element.remove()
    win.element.append(ct.element)
    if not closable:
        ct.close.style.display = 'none'
    return WindowResult(win)


class ClosableTitle(wpc.Component):
    title: js.HTMLElement = wpc.element()
    close: js.HTMLElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = f"""
<div style="display: flex; justify-content: center; align-items: center;">
    <span data-name='title' style="flex-grow: 1; text-align: center;"></span>&nbsp;
    <button data-name="close" style="cursor:pointer;">X</button>
</div> """



from wwwpy.common.designer.element_library import AttributeDef, Help, EventDef

ad_value = AttributeDef('value', Help('The value of the element.', ''))
ad_disabled = AttributeDef('disabled', Help('Whether the control is disabled.', ''), boolean=True)
ad_readonly = AttributeDef('readonly', Help('Whether the control is read-only.', ''), boolean=True)
ad_autofocus = AttributeDef('autofocus',
                            Help('Whether the control should have input focus when the page loads.', ''),
                            boolean=True)
ad_required = AttributeDef('required', Help('Whether the control is required for form submission.', ''),
                           boolean=True)
ad_placeholder = AttributeDef('placeholder', Help('A hint to the user of what can be entered in the field.', ''))
ad_name = AttributeDef('name', Help('Name of the control, useful for form submission.', ''))
ed_click = EventDef('click', Help('The element was clicked.',
                                  'https://developer.mozilla.org/en-US/docs/Web/API/Element/click_event'))
ed_dblclick = EventDef('dblclick', Help('The element was double-clicked.',
                                        'https://developer.mozilla.org/en-US/docs/Web/API/Element/dblclick_event'))
ed_input = EventDef('input', Help('The input event fires when the value of the element has been changed '
                                  'as a direct result of a user action',
                                  'https://developer.mozilla.org/en-US/docs/Web/API/Element/input_event'))
ed_keydown = EventDef('keydown', Help('The keydown event is fired when a key is pressed.'
                                      , 'https://developer.mozilla.org/en-US/docs/Web/API/Element/keydown_event'))
ed_change = EventDef('change', Help(
    'The change event is fired when a change to the element\'s value is committed by the user.',
    'https://developer.mozilla.org/en-US/docs/Web/API/Element/change_event'))



from __future__ import annotations

import asyncio

import js

import wwwpy.remote.component as wpc
from wwwpy.common import files
from wwwpy.remote import dict_to_js, dict_to_py
from wwwpy.server.designer import rpc
from . import quickstart_ui
from .quickstart_ui import QuickstartUI
from .toolbox import ToolboxComponent


def show():
    js.document.body.append(DevModeComponent().element)


class classproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


class DevModeComponent(wpc.Component, tag_name='wwwpy-dev-mode-component'):
    toolbox: ToolboxComponent = wpc.element()
    quickstart: QuickstartUI | None = None

    @classproperty
    def instance(cls) -> DevModeComponent:  # noqa
        element = js.document.getElementsByTagName(DevModeComponent.component_metadata.tag_name)
        if element.length == 0:
            raise Exception('DevModeComponent is not available')
        first_element = element[0]
        component = wpc.get_component(first_element)
        if not isinstance(component, DevModeComponent):
            raise Exception(f'element is not a DevModeComponent: {dict_to_py(first_element)}')
        return component

    def root_element(self):
        return self.shadow

    def init_component(self):
        self.shadow = self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.shadow.innerHTML = """
<wwwpy-toolbox data-name="toolbox"></wwwpy-toolbox>        
        """

        async def check_for_quickstart():
            if await rpc.quickstart_possible():
                self.quickstart = quickstart_ui.create()
                self.shadow.append(self.quickstart.window.element)
                self.toolbox.visible = False

        asyncio.create_task(check_for_quickstart())



from __future__ import annotations

from abc import ABC, abstractmethod
from time import sleep

from wwwpy.http import HttpRoute
from wwwpy.server.wait_url import wait_url
from wwwpy.websocket import WebsocketRoute


class Webserver(ABC):
    def __init__(self) -> None:
        self.host: str = '0.0.0.0'
        self.port: int = 7777

    def set_host(self, host: str) -> 'Webserver':
        self.host = host
        return self

    def set_port(self, port: int) -> 'Webserver':
        self.port = port
        return self

    def start_listen(self) -> 'Webserver':
        indent = '    '
        device_ip = f'{indent}http://{_get_ip()}:{self.port}\n' if self.host == '0.0.0.0' else ''
        print(f'Available at (non-exhaustive list):\n' +
              device_ip +
              f'{indent}{self.localhost_url()}\n'
              )
        self._start_listen()
        self.wait_ready()
        return self

    def set_http_route(self, *http_routes: HttpRoute | WebsocketRoute) -> 'Webserver':
        for http_route in http_routes:
            self._setup_route(http_route)
        return self

    @abstractmethod
    def _setup_route(self, route: HttpRoute | WebsocketRoute) -> None:
        pass

    @abstractmethod
    def _start_listen(self) -> None:
        pass

    def wait_ready(self) -> 'Webserver':
        wait_url(self.localhost_url() + '/check_if_webserver_is_accepting_requests')
        return self

    def localhost_url(self) -> str:
        return f'http://127.0.0.1:{self.port}'


def wait_forever() -> None:
    while True:
        sleep(10)


def _get_ip():
    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(('10.254.254.254', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP



from typing import List

from wwwpy.common.designer.el_common import ad_value, ad_disabled, ad_readonly, ad_autofocus, ad_required, \
    ad_placeholder, ad_name, ed_click, ed_dblclick, ed_input, ed_keydown, ed_change
from wwwpy.common.designer.element_library import ElementDef, Help, AttributeDef, NamedListMap


def _standard_elements_def() -> List[ElementDef]:
    res = [
        ElementDef(
            'button', 'js.HTMLButtonElement',
            help=Help('A clickable button.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/button'),
            attributes=[
                AttributeDef('type', Help('The type of the button.', ''),
                             values=['submit', 'reset', 'button'], default_value='button'),
                ad_value,
            ],
            events=[ed_click, ed_dblclick, ed_keydown],
        ),
        ElementDef(
            'input', 'js.HTMLInputElement',
            help=Help('A field for entering text.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input'),
            attributes=[
                AttributeDef('type', Help('The type of control to display. The default type is text.', ''),
                             values=['text', 'password', 'checkbox', 'radio', 'button', 'submit', 'reset', 'file',
                                     'hidden', 'image', 'date', 'datetime-local', 'month', 'time', 'week',
                                     'number', 'range', 'email', 'url', 'search', 'tel', 'color'],
                             default_value='text'),
                ad_value,
                ad_placeholder,
                ad_disabled,
                ad_readonly,
                ad_required,
                AttributeDef('min', Help('The minimum value of the input field.', '')),
                AttributeDef('max', Help('The maximum value of the input field.', '')),
                AttributeDef('step', Help('The legal number intervals for the input field.', '')),
                AttributeDef('pattern', Help('A regular expression that the input\'s value is checked against.', '')),
                AttributeDef('autocomplete', Help('Whether the control is required for form submission.', ''),
                             values=['on', 'off']),
                ad_autofocus,
                AttributeDef('checked', Help('Whether the control is checked.', ''), boolean=True),
                AttributeDef('multiple', Help('Whether the user is allowed to enter more than one value.', ''),
                             boolean=True),

            ],
            events=[ed_input, ed_click, ed_keydown]
        ),
        ElementDef(
            'textarea', 'js.HTMLTextAreaElement',
            help=Help(
                'A multi-line text input control.',
                'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/textarea'
            ),
            attributes=[
                AttributeDef(
                    'cols',
                    Help('The visible width of the text control, in average character widths.', ''),
                    default_value='20'
                ),
                AttributeDef(
                    'rows',
                    Help('The number of visible text lines for the control.', ''),
                    default_value='2'
                ),
                ad_disabled,
                ad_readonly,
                ad_required,
                ad_placeholder,
                ad_autofocus,
                AttributeDef(
                    'maxlength',
                    Help('The maximum number of characters that the user can enter.', '')
                ),
                AttributeDef(
                    'minlength',
                    Help('The minimum number of characters that the user can enter.', '')
                ),
                AttributeDef(
                    'wrap',
                    Help('How the control wraps text.', ''),
                    values=['soft', 'hard', 'off'],
                    default_value='soft'
                ),
                AttributeDef(
                    'autocomplete',
                    Help('Whether the control has autocomplete enabled.', ''),
                    values=['on', 'off'],
                    default_value='on'
                ),
                AttributeDef(
                    'spellcheck',
                    Help('Whether spell checking is enabled for the control.', ''),
                    values=['true', 'false'],
                    default_value='true'
                ),
                AttributeDef(
                    'form',
                    Help('The form element that the textarea is associated with (its id).', '')
                ),
                ad_name,
            ],
            events=[ed_input, ed_change, ed_keydown],
        ),
        ElementDef(
            'div', 'js.HTMLDivElement',
            help=Help('A generic container element.',
                      'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/div')
        ),
        ElementDef(
            'progress', 'js.HTMLProgressElement',
            help=Help('A progress bar.',
                      'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/progress'),
            attributes=[
                AttributeDef('value', Help('The current value of the progress bar.', '')),
                AttributeDef('max', Help('The maximum value of the progress bar.', '')),
            ],
            events=[ed_click, ]
        ),
        ElementDef(
            'select', 'js.HTMLSelectElement',
            help=Help(
                'A control that provides a menu of options.',
                'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/select'
            ),
            attributes=[
                ad_name,
                ad_disabled,
                AttributeDef(
                    'multiple',
                    Help('Whether multiple options can be selected.', ''),
                    boolean=True
                ),
                ad_required,
                AttributeDef(
                    'size',
                    Help('Number of visible options.', ''),
                    default_value='1'
                ),
                ad_autofocus,
                AttributeDef(
                    'form',
                    Help('The form element that the select is associated with (its id).', '')
                ),
            ],
            events=[ed_change, ed_input, ed_keydown],

        ),
        ElementDef(
            'meter', 'js.HTMLMeterElement',
            help=Help(
                'A scalar measurement within a known range (e.g., disk usage, battery level).',
                'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/meter'
            ),
            attributes=NamedListMap([
                AttributeDef('value', Help('The current numeric value.', '')),
                AttributeDef('min', Help('The lower numeric bound of the measured range. Default is 0.', '')),
                AttributeDef('max', Help('The upper numeric bound of the measured range. Default is 1.', '')),
                AttributeDef('low', Help('The upper numeric bound of the low end of the measured range.', '')),
                AttributeDef('high', Help('The lower numeric bound of the high end of the measured range.', '')),
                AttributeDef('optimum', Help('The optimal numeric value.', '')),
                AttributeDef('form', Help('Associates the meter with a form element.', '')), ]
            ),
            events=NamedListMap([ed_click, ]),
        )

    ]

    for r in res:
        r.gen_html = _generateHtml
    return res


def _generateHtml(element_def: ElementDef, name: str) -> str:
    tag_name = element_def.tag_name

    def _def(placeHolder=False, add='', inner=''):
        def inner_func():
            pl = '' if not placeHolder else f' placeholder="{name}"'
            add1 = '' if not add else f' {add}'
            content = name if not inner else inner
            return f'<{tag_name} data-name="{name}"{pl}{add1}>{content}</{tag_name}>'

        return inner_func

    func = {
        'button': _def(),
        'div': _def(),
        'input': lambda: f'<input data-name="{name}" placeholder="{name}">',
        'progress': lambda: f'<progress data-name="{name}" value="70" max="100">70%</progress>',
        'textarea': _def(placeHolder=True),
        'select': _def(inner='''
            <option value="option1">Option 1</option>
            <option value="option2">Option 2</option>
            <option value="option3">Option 3</option>
        '''),
        'meter': _def(add='value="0.85" min="0" max="1" low="0.3" high="0.7" optimum="0.8"')
    }
    gen_html = func.get(tag_name, None)
    html = '\n' + gen_html() if gen_html else '' + ElementDef.default_gen_html(element_def, name)
    return html

# [
#     _comp('input', 'js.HTMLInputElement', '<input type="text" data-name="#name#" value="#name#">'),
#     _comp('button', 'js.HTMLButtonElement', '<button data-name="#name#">#name#</button>'),
#     _comp('textarea', 'js.HTMLTextAreaElement', '<textarea data-name="#name#">#name#</textarea>'),
#     _comp('select', 'js.HTMLSelectElement',
#           '<select data-name="#name#"><option>#name#</option></select>'),
#     _comp('div', 'js.HTMLDivElement', '<div data-name="#name#">#name#</div>'),
#     _comp('p', 'js.HTMLParagraphElement', '<p data-name="#name#">#name#</p>'),
#     _comp('h1', 'js.HTMLHeadingElement', '<h1 data-name="#name#">#name#</h1>'),
#     _comp('h2', 'js.HTMLHeadingElement', '<h2 data-name="#name#">#name#</h2>'),
#     _comp('h3', 'js.HTMLHeadingElement', '<h3 data-name="#name#">#name#</h3>'),
#     _comp('a', 'js.HTMLAnchorElement', '<a href="#" data-name="#name#">#name#</a>'),
#     _comp('img', 'js.HTMLImageElement', '<img src="#" alt="#name#" data-name="#name#">'),
#     _comp('ul', 'js.HTMLUListElement', '<ul data-name="#name#"><li>#name#</li></ul>'),
#     _comp('ol', 'js.HTMLOListElement', '<ol data-name="#name#"><li>#name#</li></ol>'),
#     _comp('li', 'js.HTMLLIElement', '<li data-name="#name#">#name#</li>'),
#     _comp('table', 'js.HTMLTableElement', '<table data-name="#name#"><tr><td>#name#</td></tr></table>'),
#     _comp('form', 'js.HTMLFormElement', '<form data-name="#name#">#name#</form>'),
#     _comp('label', 'js.HTMLLabelElement', '<label data-name="#name#">#name#</label>'),
#
# ]



_no_remote_infrastructure_found_text = """
Import of package "remote" failed. Unable to bootstrap the application."""

# language=html
_remote_module_not_found_html = """
Warning, no module 'remote' was found.
<br>
<br>
This may be because the running directory is not a valid wwwpy project directory.
<br>
<br>
If you want to create a new project from the quickstarter, be sure to run from an <i>empty</i> directory.
<br>
Read for more information about the 
 <a href="https://github.com/wwwpy-labs/wwwpy/blob/main/docs/introduction.md#quickstart-in-dev-mode" target='_blank'>
 quickstarter</a> and a 
 <a href="https://github.com/wwwpy-labs/wwwpy/blob/main/docs/introduction.md#a-wwwpy-project-structure" target='_blank'>
 general project structure.</a>
"""
_warn = '\033[33m[WARN]\033[0m'
_remote_module_not_found_console = f"""{_warn} The current project folder is not empty: $[directory]  
{_warn} It does not contain a `remote` package, which is required for a wwwpy project.  
If you want to create a new project from the quickstarter, be sure to run from an empty directory.
Read for more information here: https://github.com/wwwpy-labs/wwwpy/blob/main/docs/introduction.md"""



import os
import logging

async def main(rootpath, invocation_dir, args):
    from wwwpy.common.designer import log_emit
    log_emit.add_once(print)
    log_emit.warning_to_log()
    logging.getLogger('wwwpy').setLevel(logging.DEBUG)

    from js import console
    console.log(f'main({rootpath}, {invocation_dir}, {args})')
    from wwwpy.common.tree import print_tree
    print_tree('/wwwpy_bundle')

# Path('/wwwpy_bundle/pytest.ini').write_text("[pytest]\n"
    #                                             "asyncio_mode = auto")
    from wwwpy.remote import micropip_install
    await micropip_install('pytest==7.2.2')  # didn't work with update to 8.1.1
    await micropip_install('pytest-asyncio')
    await micropip_install('pytest-xvirt')
    import wwwpy.common.designer as des
    for package in des.pypi_packages:
        await micropip_install(package)
    import pytest
    print('-=-' * 20 + 'pytest imported')

    os.chdir(invocation_dir)
    pytest.main(args)



from __future__ import annotations

from dataclasses import dataclass


@dataclass()
class ClassPath:
    class_module: str
    """The module name of the Component."""
    class_name: str
    """The class name of the Component."""



import asyncio

_active_tasks = set()

def create_task_safe(coro):
    """Create an asyncio task and keep a reference to prevent garbage collection.

    See https://stackoverflow.com/questions/71938799/python-asyncio-create-task-really-need-to-keep-a-reference
    """
    task = asyncio.create_task(coro)
    _active_tasks.add(task)
    task.add_done_callback(_active_tasks.discard)
    return task


from typing import NamedTuple, Callable, Union


class HttpRequest(NamedTuple):
    method: str
    content: Union[str, bytes]
    content_type: str


class HttpResponse(NamedTuple):
    content: Union[str, bytes]
    content_type: str

    @staticmethod
    def application_zip(content: bytes) -> 'HttpResponse':
        content_type = 'application/zip, application/octet-stream, application/x-zip-compressed, multipart/x-zip'
        return HttpResponse(content, content_type)

    @staticmethod
    def text_html(content: str) -> 'HttpResponse':
        return HttpResponse(content, 'text/html')


class HttpRoute(NamedTuple):
    path: str
    callback: Callable[[HttpRequest], HttpResponse]



import js
from pyodide.ffi import create_proxy
import asyncio
from wwwpy.common import quickstart
from wwwpy.remote.designer.ui.window_component import new_window, DraggableComponent
from wwwpy.remote.designer.ui.searchable_list_1 import SearchableList1, Item
from wwwpy.server.designer import rpc


class QuickstartUI:

    def __init__(self):
        self.window: DraggableComponent = new_window("Select a quickstart", closable=False).window
        # self.window.set_size('300px', '300px')
        self.window.set_position('5px', '5px')
        cmp1 = SearchableList1()
        cmp1.root_element().style.color = 'white'
        quickstart_list = quickstart.quickstart_list()
        self.quickstart_list = quickstart_list
        cmp1.items = [Item(qs.title, qs.description, {'quickstart': qs}) for qs in quickstart_list]
        cmp1.placeholder = 'Search or select below...'
        self.window.element.append(cmp1.element)

        cmp1.element.addEventListener('item-click', create_proxy(self._item_click_handler))

    def accept_quickstart(self, quickstart_name: str):
        self.window.element.remove()

        async def _notify_server():
            await rpc.quickstart_apply(quickstart_name)

        asyncio.create_task(_notify_server())

    def _item_click_handler(self, event):
        item = event.detail
        assert isinstance(item, Item)
        self.accept_quickstart(item.values['quickstart'].name)


def create() -> QuickstartUI:
    return QuickstartUI()



from __future__ import annotations

from pathlib import Path
from typing import List, Set

from wwwpy.common.filesystem import sync


def filter_by_directory(events: List[sync.Event], directory_set: Set[str | Path]) -> List[sync.Event]:
    """Rebase events that falls in directory_set.
    It will retain the original path because the events will be used to send sync content to the remote;
    so the files still need to be accessible."""
    new_events = []
    path_set = {Path(d) for d in directory_set}

    def accept(event: sync.Event) -> bool:
        src_path = Path(event.src_path)
        for path in path_set:
            if src_path.is_relative_to(path):
                return True
        return False

    for event in events:
        if accept(event):
            new_events.append(event)
    return new_events



import textwrap
from typing import List, Iterable, Tuple

from wwwpy.http import HttpRoute, HttpResponse
from wwwpy.resources import Resource, build_archive, ResourceIterable

bootstrap_javascript_placeholder = '// #bootstrap-placeholder#'


def bootstrap_routes(
        resources: List[ResourceIterable],
        python: str,
        zip_route_path: str = '/wwwpy/bundle.zip',
        extract_dir: str = None,
        html: str = f'<!DOCTYPE html><h1>Loading...</h1><script>{bootstrap_javascript_placeholder}</script>',
) -> Tuple[HttpRoute, HttpRoute]:
    """Returns a tuple of two routes: (bootstrap_route, zip_route)"""
    if extract_dir is None:
        from wwwpy.common import files
        extract_dir = files._bundle_path

    def zip_response() -> HttpResponse:
        from itertools import chain
        zip_bytes = build_archive(chain.from_iterable(resources))
        return HttpResponse.application_zip(zip_bytes)

    zip_route = HttpRoute(zip_route_path, lambda request: zip_response())

    bootstrap_python = f"""
import sys
from pyodide.http import pyfetch
response = await pyfetch('{zip_route.path}')
await response.unpack_archive(extract_dir='{extract_dir}')
sys.path.insert(0, '{extract_dir}')

{python}
    """

    javascript = get_javascript_for(bootstrap_python)
    html_replaced = html.replace(bootstrap_javascript_placeholder, javascript)
    bootstrap_route = HttpRoute('/', lambda request: HttpResponse.text_html(html_replaced))
    return bootstrap_route, zip_route


def get_javascript_for(python_code: str) -> str:
    return _js_content.replace('# python replace marker', python_code)


# language=javascript
_js_content = """
if (typeof loadPyodide === 'undefined') {
    console.log('loading pyodide...');
    let script = document.createElement('script');
    script.src = 'https://cdn.jsdelivr.net/pyodide/v0.26.4/full/pyodide.js';
    script.onload = async () => {
        let pyodide = await loadPyodide();
        window.pyodide = pyodide;
        console.log('loading pyodide.runPythonAsync(...). See in the following lines for the code');
        console.log('-----------------------  START PYTHON CODE  -------------------------------');
        console.log(`# python replace marker`);
        console.log('-----------------------  END PYTHON CODE    -------------------------------');
        pyodide.runPythonAsync(`# python replace marker`);
    };
    document.body.append(script)
}
"""


def wrap_in_tryexcept(code: str, exception_block: str) -> str:
    """It will wrap the code in try/except and catch `Exception as exception`"""
    result = 'try:\n' + textwrap.indent(code, ' ' * 4) + '\n' + \
             'except Exception as exception:\n' + textwrap.indent(exception_block, ' ' * 4)
    return result



import importlib
import importlib.abc
import importlib.util
import sys
import ast
from typing import Set

from wwwpy.common.rpc import func_registry
from ast import Module, FunctionDef, AsyncFunctionDef, ClassDef


class CustomLoader(importlib.abc.Loader):
    def __init__(self, loader):
        self.loader = loader

    def create_module(self, spec):
        return None

    def exec_module(self, module):
        module_name = module.__name__
        source = self.loader.get_source(module_name)
        proxy_source = func_registry.source_to_proxy(module_name, source)
        code = compile(proxy_source, module.__file__, 'exec')
        exec(code, module.__dict__)


class CustomFinder(importlib.abc.MetaPathFinder):
    """It intercepts the packages specified and rewrites them as a proxy-to-remote.
    The goal is to seamlessly invoke remote functions from the server"""

    def __init__(self, packages_name: Set[str], custom_loader=CustomLoader):
        self.custom_loader = custom_loader
        self.packages_name = packages_name
        super().__init__()

    def find_spec(self, fullname, path, target=None):
        if fullname in self.packages_name:
            # Temporarily remove this finder from sys.meta_path to avoid recursion. Remove also pytest rewriter hook
            orig = sys.meta_path.copy()
            sys.meta_path = [f for f in sys.meta_path if not isinstance(f, CustomFinder)]
            sys.meta_path = [f for f in sys.meta_path if f.__class__.__name__ != 'AssertionRewritingHook']
            try:
                spec = importlib.util.find_spec(fullname)
            finally:
                # Reinsert this finder back into sys.meta_path
                sys.meta_path = orig

            if spec:
                spec.loader = self.custom_loader(spec.loader)
            return spec
        return None



from __future__ import annotations

import asyncio
import logging

import js

logger = logging.getLogger(__name__)

_wwwpy_dev_mode = 'wwwpy_dev_mode'


async def activate():
    setattr(js.window, _wwwpy_dev_mode, True)
    from wwwpy.remote import micropip_install
    from wwwpy.common import designer

    for package in designer.pypi_packages:
        await micropip_install(package)

    from wwwpy.remote.designer import helpers
    js.window.onerror = helpers._on_error
    js.window.onunhandledrejection = helpers._on_unhandledrejection
    from wwwpy.remote.designer import log_redirect
    log_redirect.redirect_logging()

    loop = asyncio.get_event_loop()
    loop.set_exception_handler(_global_exception_handler)


def is_active() -> bool:
    return hasattr(js.window, _wwwpy_dev_mode)


def _global_exception_handler(loop, context):
    # The context parameter contains details about the exception
    logger.info(f"Global handler caught: {context['message']}")
    exception = context.get('exception')
    if exception:
        logger.info(f"Exception type: {type(exception)}, Args: {exception.args}")
        logger.exception(exception)



from __future__ import annotations

from dataclasses import dataclass
from typing import Protocol, Callable

from js import document, MouseEvent, HTMLElement
import js
from pyodide.ffi import create_proxy

from wwwpy.common.designer.html_edit import Position


@dataclass
class DropZone:
    element: HTMLElement
    position: Position

class DropZoneHover(Protocol):
    """The protocol for the callback function that will be called when the user hovers in/out of a drop zone."""
    def __call__(self, event: DropZone | None) -> None: ...

def _default_extract_target(event: MouseEvent) -> HTMLElement:
    return event.target


def _extract_first_with_data_name(event: MouseEvent) -> HTMLElement:
    ele: js.HTMLElement = event.target
    while ele is not None:
        if ele.hasAttribute('data-name'):
            return ele
        ele = ele.parentElement
    return event.target


# todo refactor to better integrate with toolbox
class _DropZoneSelector:
    def __init__(self, ):
        self._last_zone: DropZone | None = None
        self._accept = None
        self._callback: DropZoneHover | None = None
        self._mousemove_proxy = create_proxy(self._mousemove)
        # todo not under test
        self._whole = False
        self._extract_target = None

    def start_selector(self, callback: DropZoneHover = None,
                       accept: Callable[[DropZone], bool] = None,
                       whole=False,
                       extract_target: Callable[[MouseEvent], HTMLElement] = _extract_first_with_data_name):
        """It starts the process for the user to select a drop zone.
        While moving the mouse, it highlights the drop zones.
        """
        self._callback = callback
        self._accept = accept
        self._whole = whole
        self._extract_target = extract_target
        _ensure_drop_zone_style()
        document.addEventListener('mousemove', self._mousemove_proxy)

    def stop(self) -> DropZone | None:
        """It stops the process of selecting a drop zone.
        It will remove the highlights and the event listener.
        It will return the selected DropZone
        """
        document.removeEventListener('mousemove', self._mousemove_proxy)
        self._remove_marker()
        result = self._last_zone
        return result

    def _mousemove(self, event: MouseEvent):
        position = _calc_position(event)
        element: HTMLElement = self._extract_target(event)
        zone_event = DropZone(element, position)
        accept = True
        if self._accept:
            accept = self._accept(zone_event)
        if not accept:
            zone_event = None

        if self._last_zone != zone_event:
            self._remove_marker()
            self._last_zone = zone_event
            if zone_event is not None:
                # console.log(f'candidate sending zone_event', zone_event.position, zone_event.target)
                if self._whole:
                    element.classList.add(_whole_css_class)
                else:
                    css_class = _beforebegin_css_class if position == Position.beforebegin else _afterend_css_class
                    element.classList.add(css_class)
            if self._callback:
                self._callback(zone_event)
        else:
            pass
            # console.log(f'candidate discarded zone_event: {zone_event}')

    def _remove_marker(self):
        if self._last_zone is not None:
            _remove_class(self._last_zone.element, _beforebegin_css_class)
            _remove_class(self._last_zone.element, _afterend_css_class)
            _remove_class(self._last_zone.element, _whole_css_class)


drop_zone_selector = _DropZoneSelector()

_beforebegin_css_class = 'drop-zone-beforebegin'
_afterend_css_class = 'drop-zone-afterend'
_whole_css_class = 'drop-zone-whole'


def _remove_class(target: HTMLElement, class_name: str):
    target.classList.remove(class_name)
    if target.classList.length == 0:
        target.removeAttribute('class')


def _ensure_drop_zone_style():
    drop_zone_highlight_style = 'drop-zone-highlight-style'
    _ensure_style_element(drop_zone_highlight_style, f"""
    .{_beforebegin_css_class} {{
        box-shadow: -4px -4px 0 green;
        position: relative;
        z-index: 10;
    }}
    .{_afterend_css_class} {{
        box-shadow: 2px 2px 2px 2px green;
        position: relative;
        z-index: 10;
    }}
    .{_whole_css_class} {{
        box-shadow: 0 0 0 2px green;
        position: relative;
        z-index: 10;
    }}
""")


def _ensure_style_element(element_id: str, style_content: str):
    st = document.head.querySelector(f'#{element_id}')
    if st is None:
        st = document.createElement('style')
        st.id = element_id
        document.head.appendChild(st)

    if st.innerHTML != style_content:
        st.innerHTML = style_content


def _calc_position(event: MouseEvent) -> Position:
    element: HTMLElement = event.target
    rect = element.getBoundingClientRect()

    # Calculate the position of the mouse relative to the element
    x = event.clientX - rect.left
    y = rect.height - (event.clientY - rect.top)
    h = rect.height
    w = rect.width
    if w == 0:
        return Position.inside
    m = h / w
    y2 = m * x
    res = Position.afterend if y <= y2 else Position.beforebegin
    # console.log(f'x={x}, y={y}, w={w}, h={h} m={m}, y2={y2} res={res}')
    return res



import logging
from datetime import timedelta
from pathlib import Path
from typing import List, Callable, Set

from wwwpy.common.files import extension_blacklist, directory_blacklist
from wwwpy.common.filesystem import sync
from wwwpy.common.filesystem.sync import sync_delta2, Sync, event_rebase
from wwwpy.common.quickstart import is_empty_project
from wwwpy.common.reloader import reload
from wwwpy.remote.designer.rpc import DesignerRpc
from wwwpy.rpc import RpcRoute
from wwwpy.server.filesystem_sync.watchdog_debouncer import WatchdogDebouncer
from wwwpy.websocket import WebsocketPool, PoolEvent

logger = logging.getLogger(__name__)

sync_impl: Sync = sync_delta2


def _warning_on_multiple_clients(websocket_pool: WebsocketPool):
    _in_warning = False

    def pool_before_change(event: PoolEvent):
        nonlocal _in_warning
        client_count = len(websocket_pool.clients)
        if client_count > 1:
            _in_warning = True
            logger.warning(f'Total browser connected: {client_count}')
        else:
            if _in_warning:
                _in_warning = False
                logger.warning(f'Back to only {client_count} browser connected')

    websocket_pool.on_after_change.append(pool_before_change)


def start_hotreload(directory: Path, websocket_pool: WebsocketPool, rpc_route: RpcRoute,
                    server_folders: Set[str], remote_folders: Set[str]):
    remote_set = {directory / d for d in remote_folders}
    server_set = {directory / d for d in server_folders}
    rpc_set = {directory / (d.replace('.', '/') + '.py') for d in rpc_route._allowed_modules}

    def process_events(events: List[sync.Event]):
        remote_events = event_rebase.filter_by_directory(events, remote_set)
        server_events = event_rebase.filter_by_directory(events, server_set)

        if len(server_events) > 0:
            _print_events('server', server_events, directory)
            do_unload_for(directory, server_folders)
            rpc_events = event_rebase.filter_by_directory(server_events, rpc_set)
            if len(rpc_events) > 0:
                # if the signature of the rpc changes, it will write on the fs and will trigger the observer
                rpc_stub_files = rpc_route.generate_remote_stubs()
                evs = [
                          sync.Event('created', True, str(rpc_route.tmp_bundle_folder / 'server')),
                          sync.Event('created', True, str(rpc_route.tmp_bundle_folder / 'server/rpc')),
                      ] + \
                      [sync.Event('modified', False, str(f)) for f in rpc_stub_files]

                _print_events('server-rpc', evs, rpc_route.tmp_bundle_folder)
                process_remote_events(rpc_route.tmp_bundle_folder, websocket_pool, evs, len(remote_events) == 0)

        if len(remote_events) > 0:
            _print_events('remote', remote_events, directory)
            process_remote_events(directory, websocket_pool, remote_events, do_reload=True)

        if len(remote_events) == 0 and len(server_events) == 0 and is_empty_project(directory):
            for remote_rpc in websocket_pool.all_clients_rpc(DesignerRpc):
                remote_rpc.hotreload_do()

    _watch_filesystem_change(directory, process_events)


def process_remote_events(directory: Path, websocket_pool: WebsocketPool, events: List[sync.Event], do_reload: bool):
    try:
        payload = sync_impl.sync_source(directory, events)
        for remote_rpc in websocket_pool.all_clients_rpc(DesignerRpc):
            remote_rpc.hotreload_notify_changes(do_reload, payload)
    except:
        # we could send a sync_init
        import traceback
        logger.error(f'_on_remote_events 1 {traceback.format_exc()}')


def do_unload_for(directory, server_folders: Set[str]):
    for p in server_folders:
        package_directory = directory / p
        if package_directory:
            try:
                import wwwpy.common.reloader as reloader
                reloader.unload_path(str(package_directory))
            except:
                # we could send a sync_init
                import traceback
                logger.error(f'_hotreload_server {traceback.format_exc()}')


def _watch_filesystem_change(directory: Path, callback: Callable[[List[sync.Event]], None]):
    def on_sync_events(events: List[sync.Event]):
        try:
            # oh, boy. When a .py file is saved it fires the first hot reload. Then, when that file is loaded
            # the python updates the __pycache__ files, firing another (unwanted) reload: the first was enough!
            filtered_events = _filter_events(events, directory)
            if len(filtered_events) > 0:
                callback(filtered_events)
        except:
            import traceback
            logger.error(f'_watch_filesystem_change {traceback.format_exc()}')

    handler = WatchdogDebouncer(directory, timedelta(milliseconds=100), on_sync_events)
    handler.watch_directory()


def _filter_events(events: List[sync.Event], directory: Path) -> List[sync.Event]:
    def reject(e: sync.Event) -> bool:
        src_path = Path(e.src_path)
        if src_path.suffix in extension_blacklist:
            return True
        p = src_path.relative_to(directory)
        for part in p.parts:
            if part in directory_blacklist:
                return True
        return False

    result = [e for e in events if not reject(e)]
    return result


def _print_events(label: str, events: List[sync.Event], root_dir: Path):
    def accept(e: sync.Event) -> bool:
        bad = e.is_directory and e.event_type == 'modified'
        return not bad

    def to_str(e: sync.Event) -> str:
        def rel(path: str) -> str:
            return str(Path(path).relative_to(root_dir))

        dest_path = rel(e.dest_path) if e.dest_path else ''
        src_path = rel(e.src_path)
        return src_path if dest_path == '' else f'{src_path} -> {dest_path}'

    summary = list(set(to_str(e) for e in events if accept(e)))
    logger.info(f'Hotreload {label} events: {len(events)}. Changes summary: {summary}')
